<?php
namespace nur\sery\wip\schema\_scalar;

use nulib\php\func;
use nulib\ref\schema\ref_analyze;
use nulib\ValueException;
use nur\sery\wip\schema\AnalyzerContext;
use nur\sery\wip\schema\input\Input;
use nur\sery\wip\schema\types;
use nur\sery\wip\schema\types\IType;
use nur\sery\wip\schema\Wrapper;

class ScalarWrapper extends Wrapper {
  function __construct(ScalarSchema $schema, &$dest=null, $destKey=null, ?array $params=null) {
    $defaultVerifix = $params["verifix"] ?? true;
    $defaultThrow = $params["throw"] ?? null;
    if ($dest !== null && $defaultThrow === null) {
      # Si $dest est null, ne pas lancer d'exception, parce qu'on considère que
      # c'est une initialisation sans conséquences
      $defaultThrow = true;
    }
    $this->schema = $schema;
    $this->verifix = $defaultVerifix;
    $this->throw = $defaultThrow ?? false;
    $this->result = new ScalarResult();
    $this->reset($dest, $destKey);
    $this->throw = $defaultThrow ?? true;
  }

  function isScalar(?ScalarWrapper &$scalar=null): bool { $scalar = $this; return true; }

  /** schéma de cette valeur */
  protected ScalarSchema $schema;

  /** source et destination de la valeur */
  protected Input $input;

  /** @var string|int|null clé de la valeur dans le tableau destination */
  protected $destKey;

  protected bool $verifix;

  protected bool $throw;

  /** type de la valeur après analyse */
  protected ?IType $type;

  /** résultat de l'analyse de la valeur */
  protected ScalarResult $result;

  function reset(&$dest, $destKey=null, ?bool $verifix=null): Wrapper {
    if ($dest instanceof Input) $input = $dest;
    else $input = new Input($dest);
    $this->input = $input;
    $this->destKey = $destKey;
    $this->type = null;
    $this->analyzeExtractParse();
    if ($verifix ?? $this->verifix) $this->verifix();
    return $this;
  }

  function getKeys(): array {
    return [null];
  }

  function getWrapper($key=null): ScalarWrapper {
    if ($key === null) return $this;
    throw ValueException::invalid_key($key);
  }

  /** analyser la valeur et résoudre son type */
  function _analyze(AnalyzerContext $context): int {
    $schema = $this->schema;
    $input = $this->input;
    $destKey = $this->destKey;
    $result = $this->result;

    if (!$input->isPresent($destKey)) return $result->setMissing($schema);

    $haveType = false;
    $types = [];
    $type = $firstType = null;
    $haveValue = false;
    $value = null;
    # d'abord chercher un type pour lequel c'est une valeur normalisée
    $index = 0;
    foreach ($schema->type as $key => $name) {
      if ($key === $index) {
        $index++;
        $params = null;
      } else {
        $params = $name;
        $name = $key;
      }
      $type = types::get($name, $params, $this->schema->getDefinition());
      if ($firstType === null) $firstType = $type;
      $types[] = $type;
      if ($type->isAvailable($input, $destKey)) {
        if (!$haveValue) {
          $value = $input->get($destKey);
          $haveValue = true;
        }
        if ($type->isValid($value, $normalized) && $normalized) {
          $haveType = true;
          $this->type = $type;
          break;
        }
      }
    }
    if (!$haveType) {
      # ensuite chercher un type pour lequel la valeur est valide
      foreach ($types as $type) {
        if ($type->isAvailable($input, $destKey) && $type->isValid($value)) {
          $haveType = true;
          $this->type = $type;
          break;
        }
      }
    }

    # sinon prendre le premier type
    if (!$haveType) $type = $this->type = $firstType;
    $context->type = $type;
    if (!$type->isAvailable($input, $destKey)) return $result->setUnavailable($schema);

    $value = $input->get($destKey);
    $context->orig = $context->dest = $value;
    if ($type->isNull($value)) {
      return $result->setNull($schema);
    } elseif (is_string($value)) {
      return ref_analyze::STRING;
    } elseif ($type->isValid($value, $normalized)) {
      if ($normalized) return $result->setNormalized();
      else return $result->setValid();
    } else {
      return $result->setInvalid($value, $schema);
    }
  }

  function _extract(AnalyzerContext $context): bool {
    return $context->type->extract($context->dest);
  }

  function _parse(AnalyzerContext $context): bool {
    try {
      $context->dest = $context->type->parse($context->dest);
      return true;
    } catch (ValueException $e) {
      return false;
    }
  }

  function analyzeExtractParse(): int {
    $schema = $this->schema;
    $input = $this->input;
    $destKey = $this->destKey;
    $result = $this->result;
    $result->reset();
    $context = new AnalyzerContext($schema, $this, $input, $destKey, $result);

    /** @var func $analyzerFunc */
    $analyzerFunc = $schema->analyzerFunc;
    if ($analyzerFunc !== null) $what = $analyzerFunc->invoke([$context]);
    else $what = $this->_analyze($context);
    if ($what !== ref_analyze::STRING) return $what;

    /** @var func $extractorFunc */
    $extractorFunc = $schema->extractorFunc;
    if ($extractorFunc !== null) $valid = $extractorFunc->invoke([$context]);
    else $valid = $this->_extract($context);
    if (!$valid) return $result->setInvalid($context->orig, $schema);
    if ($context->type->isNull($context->dest)) return $result->setNull($schema);
    $extracted = $context->dest;

    /** @var func $parserFunc */
    $parserFunc = $schema->parserFunc;
    if ($parserFunc !== null) $valid = $parserFunc->invoke([$context]);
    else $valid = $this->_parse($context);
    if ($valid) {
      $normalized = $context->dest === $context->orig;
      if ($normalized) {
        $input->set($context->dest, $destKey);
        return $result->setNormalized();
      } else {
        $input->set($extracted, $destKey);
        return $result->setValid();
      }
    }
    return $result->setInvalid($context->orig, $schema);
  }

  function verifix(?bool $throw=null): bool {
    $schema = $this->schema;
    $input = $this->input;
    $destKey = $this->destKey;
    $result = $this->result;
    $context = new AnalyzerContext($schema, $this, $input, $destKey, $result);
    /** @var func $normalizerFunc */
    $normalizerFunc = $schema->normalizerFunc;
    if ($normalizerFunc !== null) return $normalizerFunc->invoke([$context]);

    if ($throw === null) $throw = $this->throw;
    $verifix = false;
    $modified = false;
    if ($result->resultAvailable) {
      if ($result->null) {
        # forcer la valeur null, parce que la valeur actuelle est peut-être une
        # valeur assimilée à null
        $this->input->set(null, $destKey);
      } elseif ($result->valid && !$result->normalized) {
        # normaliser la valeur
        $verifix = true;
      }
    } else {
      $verifix = true;
    }
    if ($verifix) {
      $value = $this->input->get($destKey);
      $modified = $this->type->verifix($value, $result, $this->schema);
      if ($result->valid) $this->input->set($value, $destKey);
    }
    if (!$result->valid) $result->throw($throw);
    return $modified;
  }

  function getResult(): ScalarResult {
    return $this->result;
  }

  function isPresent(): bool {
    return $this->result->present;
  }

  function getType(): IType {
    return $this->type;
  }

  function isAvailable(): bool {
    return $this->result->available;
  }

  function isValid(): bool {
    return $this->result->valid;
  }

  function isNormalized(): bool {
    return $this->result->normalized;
  }

  function get($default=null) {
    if ($this->result->available) return $this->input->get($this->destKey);
    else return $default;
  }

  function set($value, ?bool $verifix=null): ScalarWrapper {
    $this->input->set($value, $this->destKey);
    $this->analyzeExtractParse();
    if ($verifix ?? $this->verifix) $this->verifix();
    return $this;
  }

  function unset(?bool $verifix=null): ScalarWrapper {
    $this->input->unset($this->destKey);
    $this->analyzeExtractParse();
    if ($verifix ?? $this->verifix) $this->verifix();
    return $this;
  }

  function format($format=null): string {
    $value = $this->input->get($this->destKey);
    /** @var func $formatterFunc */
    $formatterFunc = $this->schema->formatterFunc;
    if ($formatterFunc !== null) {
      # la fonction formatter n'a pas forcément accès au format de la définition
      # le lui fournir ici
      $format ??= $this->schema->format;
      return $formatterFunc->invoke([$value, $format]);
    } else {
      # on assume que le type a été initialisé avec le format de la définition
      # le cas échéant
      return $this->type->format($value, $format);
    }
  }
}
